from webmother.passport import Passport
from tweb.license import License
from webmother.service import ctrl_passport

system_name = 'product'

max_lic_text = 'product:111111111;30;'
profiles = {
    'product': {
        'switch': [
            "create",
            "read",
            "update",
            "remove",
            "submit",
            "audit",
            "reject",
            "activate",
            "deactivate"
        ],
        'number': [
            "visible_level"  # 资源可见级别，越大表示可以看到status值更低的资源，取值范围为资源status取值范围，如0～40
        ],
    }
}
display = {
    'zh': {
        'product': '产品管理',
        'product.switch': '权限开关',
        'product.switch.create': '创建',
        'product.switch.read': '读取',
        'product.switch.update': '更新',
        'product.switch.remove': '移除',
        'product.switch.submit': '提交',
        'product.switch.audit': '审核',
        'product.switch.reject': '驳回',
        'product.switch.activate': '上架',
        'product.switch.deactivate': '下架',
        'product.number': '数量限制',
        'product.number.visible_level': '可见级别',
    },
    'en': {
        'product': 'Product Manage',
        'product.switch': 'Switches',
        'product.switch.create': 'Create',
        'product.switch.read': 'Read',
        'product.switch.update': 'Update',
        'product.switch.remove': 'Remove',
        'product.switch.submit': 'Submit',
        'product.switch.audit': 'Audit',
        'product.switch.reject': 'Reject',
        'product.switch.activate': 'Activate',
        'product.switch.deactivate': 'Deactivate',
        'product.number': 'Number Limit',
        'product.number.visible_level': 'Visible Level',
    }
}


def append_extra():
    # 添加本系统涉及到的权限项
    Passport.add_system_profile(system_name, profiles, display)

    # 更新超级管理员的账户相关权限
    extra_lic = License(profiles, display).parse(max_lic_text)
    ctrl_passport.simple_update('0', '0', extra_lic.json)



# coding:utf-8

from ..db import mongo as db
from tweb.json_util import filter_keys
from tweb.error_exception import ErrException, ERROR
from tweb import time
from webmother.utils.bson_util import bson_id, json_id
from webmother.service import ctrl_catalog
from webmother.passport import Passport
import re


# 分类节点状态以及状态迁移定义
class Status:
    removed = -10  # 已删除
    editing = 0  # 编辑中
    auditing = 10  # 待审(审核中)
    sleeping = 20  # 休眠中
    activated = 30  # 已激活

    default = activated

    status_map = {
        editing: {'submit': auditing, 'remove': removed},
        auditing: {'reject': editing, 'audit': sleeping, 'remove': removed},
        sleeping: {'activate': activated, 'remove': removed},
        activated: {'deactivate': sleeping}
    }

    @staticmethod
    def trans(cur_status, action):
        """
        在当前状态，进行操作将会得到新的状态
        :param cur_status: 当前状态
        :param action: 操作名称
        :return: 新的状态
        """

        valid_actions = Status.status_map.get(cur_status)
        if valid_actions is None:
            raise ErrException(ERROR.E40022, extra=f'current status is {cur_status}, forbid change status')

        new_status = valid_actions.get(action)
        if new_status is None:
            raise ErrException(ERROR.E40022, extra=f'current status is {cur_status}, wrong action [{action}]')

        return new_status


def create(cid, prod_meta, *auth_args):
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    Passport().verify(*auth_args).operable(c.get('node'), 'product.create')
    # END

    prod_meta['name'] = prod_meta['name'].lower()
    name = prod_meta['name']
    if not re.match(r'^[\w-]{0,19}$', name):
        raise ErrException(ERROR.E40000, extra='name should be 20 letter, number, _, -, and beginning with letter')

    if 'the_pledged' not in prod_meta:
        raise ErrException(ERROR.E40000, extra='no the_pledged field')
    if 'the_loaned' not in prod_meta:
        raise ErrException(ERROR.E40000, extra='no the_loaned field')
    if 'normal_rate' not in prod_meta:
        raise ErrException(ERROR.E40000, extra='no normal_rate field')
    if 'warn_rate' not in prod_meta:
        raise ErrException(ERROR.E40000, extra='no warn_rate field')
    if 'bomb_rate' not in prod_meta:
        raise ErrException(ERROR.E40000, extra='no bomb_rate field')

    if c['status'] not in [ctrl_catalog.Status.sleeping, ctrl_catalog.Status.activated]:
        raise ErrException(ERROR.E40300, extra='can not add something into catalog because it not audited')

    if db.product.find_one({'name': name, 'status': {'$gte': 0}}) is not None:
        raise ErrException(ERROR.E40020)

    prod_meta['catalog'] = bson_id(cid)
    prod_meta['status'] = Status.default

    now = time.millisecond()
    prod_meta['created'] = now
    prod_meta['updated'] = now

    result = db.product.insert_one(prod_meta)
    return simple_read(result.inserted_id)


def read(cid, pid, *auth_args):
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    pp = Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'product.read')
    visible_level = pp.number(c.get('node'), 'product.visible_level')
    min_stat = Status.activated - visible_level
    # END

    return simple_read(pid, min_stat)


def simple_read(pid, min_stat=0):
    p = db.product.find_one({'_id': bson_id(pid), 'status': {'$gte': min_stat}}, {'catalog': 0})
    if p is None:
        raise ErrException(ERROR.E40400, extra=f'the product({pid}) not existed')

    p['pid'] = json_id(p.pop('_id'))

    return p


def update(cid, pid, prod_meta, *auth_args):
    """
    修改产品信息
    :param cid: 产品所属分类节点ID
    :param pid: 产品id
    :param prod_meta: 产品信息
    :param auth_args: 鉴权参数：(signed, nonce), 即("签名的授权字符串", "临时一致性标示，需与生成签名时使用的nonce相同")
    :return:
    """
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'product.update')
    # END

    p = simple_read(pid)
    if p is None:
        raise ErrException(ERROR.E40400, extra='wrong product id')

    if p['status'] not in (Status.editing, Status.auditing, Status.sleeping, Status.activated):
        raise ErrException(ERROR.E40021)

    new_data = filter_keys(prod_meta, {
        'display': 1,
        'desc': 1,
        'icon': 1
    })

    new_data['status'] = Status.default
    new_data['updated'] = time.millisecond()

    db.product.update_one({'_id': bson_id(pid)}, {'$set': new_data})
    return simple_read(pid)


def change_status(cid, pid, action, *auth_args):
    """
    :param cid: 产品所属分类节点ID
    :param pid: 产品id
    :param action: 操作（提交，过审，驳回，上架，下架，删除等）
    :param auth_args: 鉴权参数：(signed, nonce), 即("签名的授权字符串", "临时一致性标示，需与生成签名时使用的nonce相同")
    :return:
    """
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), f'product.{action}')
    # END

    p = simple_read(pid)
    if p is None:
        raise ErrException(ERROR.E40400, extra='wrong product id')

    cur_status = p.get('status')
    new_status = Status.trans(cur_status, action)

    new_data = {
        'status': new_status,
        'updated': time.millisecond()
    }

    db.product.update_one({'_id': bson_id(pid)}, {'$set': new_data})

    return {'id': pid, 'status': new_status, 'old_status': cur_status}


def move(cid, pid, cid_to, *auth_args):
    """
    把cid节点下的pid产品移到cid_to标示的节点之下
    :param cid: 原节点ID
    :param pid: 产品ID
    :param cid_to: 新节点ID
    :param auth_args: 鉴权参数：(signed, nonce), 即("签名的授权字符串", "临时一致性标示，需与生成签名时使用的nonce相同")
    :return:
    """
    c = ctrl_catalog.simple_read(cid)
    c_to = ctrl_catalog.simple_read(cid_to)

    if cid == cid_to:
        raise ErrException(ERROR.E40000, extra='the same node, not need move')

    # 授权检查
    Passport().verify(*auth_args).operable('{}/*'.format(c['node']), 'product.remove')
    Passport().verify(*auth_args).operable('{}/*'.format(c_to['node']), 'product.create')
    # END

    p = simple_read(pid)
    if p['status'] not in (Status.editing, Status.auditing, Status.sleeping, Status.activated):
        raise ErrException(ERROR.E40021)

    now = time.millisecond()

    tmp_p = {
        'catalog': bson_id(cid_to),
        'updated': now
    }

    db.product.update_one({'_id': bson_id(pid)}, {'$set': tmp_p})

    return simple_read(pid)


def query_products(cid, *auth_args):
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    pp = Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'product.read')
    visible_level = pp.number(c.get('node'), 'product.visible_level')
    min_stat = Status.activated - visible_level
    # END

    cursor = db.product.find({'catalog': bson_id(cid), 'status': {'$gte': min_stat}}, {'catalog': 0})

    array = list()
    for item in cursor:
        item['pid'] = json_id(item.pop('_id'))
        array.append(item)

    return array



# coding:utf-8

import config
import pymongo
from tornado.options import options
from webmother.db import mongo as mother_db

MongoDB = {
    'active': True,  # 是否启用
    'dev': {
        'host': config.Mongo['host'],
        'port': config.Mongo['port'],
        'db': 'proddb',
        'user': 'app',
        'pwd': 'Prod2app',
        'connecttimeoutms': 60 * 1000
    },
    'prod': {
        'host': config.Mongo['host'],
        'port': config.Mongo['port'],
        'db': 'proddb',
        'user': 'app',
        'pwd': 'Prod2app',
        'connecttimeoutms': 60 * 1000
    }
}

# web-mother相关集合
catalog = None
org = None
org2user = None
catalog2org = None

# product DB
mongo_client = None
mongo_db = None

product = None


def init():
    if not MongoDB['active']:
        return

    global catalog
    global org

    global org2user
    global catalog2org

    global mongo_client
    global mongo_db

    global product

    mongo_cfg = MongoDB[options.env]

    if mongo_client is not None:
        return

    mother_db.init()

    # 引用webmother中的数据集合
    catalog = mother_db.catalog
    org = mother_db.org
    org2user = mother_db.org2user
    catalog2org = mother_db.catalog2org
    # END

    mongo_client = pymongo.MongoClient(host=mongo_cfg['host'],
                                       port=mongo_cfg['port'],
                                       username=mongo_cfg['user'],
                                       password=mongo_cfg['pwd'],
                                       authSource=mongo_cfg['db'])
    mongo_db = mongo_client[mongo_cfg['db']]

    # my collections
    product = mongo_db.product
    # END

    # 创建索引
    _product_index()

    # 初始化系统数据
    _init_data()


def start_session():
    return mongo_client.start_session()


def _product_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "name": "live-pledge",                           # 产品标示名，英文、数字、"-"
        "catalog": ObjectId('5c710622e155ac0c39c8b66d'), # 所属分类节点ID
        "status": 30,                                    # 状态：-10:已删除；0:编辑中；10:待审；20:休眠中；30:已激活；
        "display": {
            "zh": "活期质押",
            "en": "Live Pledge"
        },
        "desc": {
            "zh": "灵活的质押和支取",
            "en": "Pledge and draw flexibly"
        },
        "icon": "http://your.com/icon/pledge.png",
        "the_pledged": {...},   # 被质押的东西
        "the_loaned": {...},    # 被贷出的东西
        "normal_rate": 1.5,   # 正常质押率，质押率：质押资产市值/贷款货币市值
        "warn_rate": 1.1,     # 禁戒质押率
        "bomb_rate": 1.05,    # 爆仓质押率
        "interest_rate": 0.0001 # 贷款利率（日利率）
        "created": 1551015331186,
        "updated": 1551015331186
    }
    """
    product.create_index('name')
    product.create_index([('catalog', pymongo.ASCENDING), ("updated", pymongo.ASCENDING)], sparse=True)


def _init_data():
    pass



# coding:utf-8

from .. import ctrl_product
from asyncio import get_event_loop


async def create(cid, prod_meta, *auth_args):
    args = (cid, prod_meta, *auth_args)
    return await get_event_loop().run_in_executor(None, ctrl_product.create, *args)


async def read(cid, pid, *auth_args):
    args = cid, pid, *auth_args
    return await get_event_loop().run_in_executor(None, ctrl_product.read, *args)


async def update(cid, pid, prod_meta, *auth_args):
    args = cid, pid, prod_meta, *auth_args
    return await get_event_loop().run_in_executor(None, ctrl_product.update, *args)


async def change_status(cid, pid, action, *auth_args):
    args = cid, pid, action, *auth_args
    return await get_event_loop().run_in_executor(None, ctrl_product.change_status, *args)


async def move(cid, pid, cid_to, *auth_args):
    args = cid, pid, cid_to, *auth_args
    return await get_event_loop().run_in_executor(None, ctrl_product.move, *args)


async def query_products(cid, *auth_args):
    args = (cid, *auth_args)
    return await get_event_loop().run_in_executor(None, ctrl_product.query_products, *args)



# coding=utf-8

import webmother
from .db import mongo
from . import routes
from .extra_passport import append_extra


def init(app, load_routes=True):
    # 初始化webmother
    webmother.init(app)

    # 初始化本系统数据库
    mongo.init()

    # 增加本系统增加的权限需求
    append_extra()

    # 加载路由模块
    if load_routes:
        app.load_routes(routes)



# coding=utf-8

import config
from ..extra_passport import system_name
from ..http_handler import product_handler, entrance_handler

base = '{}/{}/{}'.format(config.VER, config.PLATFORM, system_name)
base_prod = '{}/(?P<fief>fief|tree|node|catalog)/(?P<cid>\w+)'.format(base)
routes = [
    # 查询组织拥有哪些资源的许可证/通行证，许可范围有多大
    (rf"/{base}/entrance/employ/passports", entrance_handler.EmployPassportsHandler),

    # 产品的增删改查
    (rf"/{base_prod}/product/(?P<pid>[a-f0-9]*)", product_handler.ProductHandler),

    # 产品的状态操作
    (rf"/{base_prod}/product/(?P<pid>[a-f0-9]*)/do/(?P<action>\w*)", product_handler.StatusHandler),

    # 将产品从一个分类中已到另一分类中
    (rf"/{base_prod}/product/(?P<pid>[a-f0-9]*)/move/(?P<cid_to>\w*)", product_handler.MovingHandler),

    # 查询分类节点下的产品列表
    (rf"/{base_prod}/list", product_handler.ListHandler),
]



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages

ver = '0.2.2'

setup(
    name='qcc-product',
    version=ver,
    description=(
        'All like QCC product manage business can from qcc-product. Web-mother include member manage, '
        'organization manage, and catalog manage. Especially qcc-product support authorization management.'
    ),
    long_description='Docs for this project are maintained at https://gitee.com/qcc100/qcc-product.git.',
    author='Yang Chunbo',
    author_email='ycb@microto.com',
    maintainer='Yang Chunbo',
    maintainer_email='ycb@microto.com',
    license='MIT License',
    packages=find_packages(),
    platforms=["all"],
    url='https://gitee.com/qcc100/qcc-product.git',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
    ]
)



# coding=utf-8

from tweb import base_handler, myweb
from tweb.error_exception import ErrException, ERROR
from tornado import gen
import json
from ..service.async_wrap import ctrl_product
from webmother.utils import user_auth


class ProductHandler(base_handler.BaseHandler):
    """
    产品基本操作：增删改查（CRUD）
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, cid, **kwargs):
        passport = self.request.headers.get('x-signed-passport')
        access_token = self.request.headers.get('x-access-token')

        data = json.loads(self.request.body.decode('utf-8'))
        if 'name' not in data:
            raise ErrException(ERROR.E40000, extra='not name field')

        ret = yield ctrl_product.create(cid, data, passport, access_token)
        return self.write(ret)

    @gen.coroutine
    def get(self, cid, pid, **kwargs):
        uid = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')
        passport = self.request.headers.get('x-signed-passport')

        auth = yield user_auth.verify(uid, access_token, self.request.remote_ip)

        ret = yield ctrl_product.read(cid, pid, passport, auth[1])
        return self.write(ret)

    @myweb.authenticated
    @gen.coroutine
    def put(self, cid, pid, **kwargs):
        passport = self.request.headers.get('x-signed-passport')
        access_token = self.request.headers.get('x-access-token')

        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield ctrl_product.update(cid, pid, data, passport, access_token)
        return self.write(ret)

    @myweb.authenticated
    @gen.coroutine
    def delete(self, cid, pid, **kwargs):
        passport = self.request.headers.get('x-signed-passport')
        access_token = self.request.headers.get('x-access-token')

        ret = yield ctrl_product.change_status(cid, pid, 'remove', passport, access_token)
        return self.write(ret)


class StatusHandler(base_handler.BaseHandler):
    """
    产品状态操作，只存在更新操作
    """

    @myweb.authenticated
    @gen.coroutine
    def put(self, cid, pid, action, **kwargs):
        passport = self.request.headers.get('x-signed-passport')
        access_token = self.request.headers.get('x-access-token')

        ret = yield ctrl_product.change_status(cid, pid, action, passport, access_token)
        return self.write(ret)


class MovingHandler(base_handler.BaseHandler):
    """
    将产品从一个分类中已到另一分类中
    """

    @myweb.authenticated
    @gen.coroutine
    def put(self, cid, pid, cid_to, **kwargs):
        passport = self.request.headers.get('x-signed-passport')
        access_token = self.request.headers.get('x-access-token')

        ret = yield ctrl_product.move(cid, pid, cid_to, passport, access_token)
        return self.write(ret)


class ListHandler(base_handler.BaseHandler):
    """
    查询目录下的产品列表
    """

    # 读取产品列表为公开接口
    @gen.coroutine
    def get(self, cid, **kwargs):
        uid = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')
        passport = self.request.headers.get('x-signed-passport')

        auth = yield user_auth.verify(uid, access_token, self.request.remote_ip)

        ret = yield ctrl_product.query_products(cid, passport, auth[1])
        return self.write({'list': ret})
